package org.bdware.consistency.plugin.raft.algo.message;

import org.bdware.consistency.plugin.raft.algo.storage.LogEntry;
import org.bdware.sc.conn.ByteUtil;
import org.bdware.sc.units.PubKeyNode;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class AppendEntries implements IRaftMessage {
    RaftMessageType type;
    long term; // leader's term
    PubKeyNode leader; // leader's id
    long prevLogIndex;
    long prevLogTerm;
    long leaderCommit; // leader's commitIndex
    List<LogEntry> entries;

    public AppendEntries() {
        this.type = RaftMessageType.AppendEntries;
    }

    public AppendEntries(long term, PubKeyNode leader, long prevLogIndex, long prevLogTerm, long leaderCommit, List<LogEntry> entries) {
        this.type = RaftMessageType.AppendEntries;
        this.term = term;
        this.leader = leader;
        this.prevLogIndex = prevLogIndex;
        this.prevLogTerm = prevLogTerm;
        this.leaderCommit = leaderCommit;
        this.entries = entries;
    }

    @Override
    public byte[] getBytes() {
        ByteArrayOutputStream msg = new ByteArrayOutputStream();
        try {
            ByteUtil.writeLong(msg, term);
            if (leader != null && !leader.pubkey.equals("")) {
                byte[] pubkey = leader.pubkey.getBytes();
                ByteUtil.writeInt(msg, pubkey.length);
                msg.write(pubkey);
            } else {
                ByteUtil.writeInt(msg, 0);
            }
            ByteUtil.writeLong(msg, prevLogIndex);
            ByteUtil.writeLong(msg, prevLogTerm);
            ByteUtil.writeLong(msg, leaderCommit);
            if (entries != null) {
                ByteUtil.writeInt(msg, entries.size());
                for (LogEntry entry : entries) {
                    byte[] entryBytes = entry.getBytes();
                    ByteUtil.writeInt(msg, entryBytes.length);
                    msg.write(entryBytes);
                }
            } else {
                ByteUtil.writeInt(msg, 0);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        return msg.toByteArray();
    }

    @Override
    public IRaftMessage parse(byte[] b) {
        AppendEntries msg = new AppendEntries();
        ByteArrayInputStream bi = new ByteArrayInputStream(b);
        msg.term = ByteUtil.readLong(bi);
        int leaderLen = ByteUtil.readInt(bi);
        PubKeyNode node = new PubKeyNode();
        if (leaderLen > 0) {
            node.pubkey = new String(ByteUtil.readBytes(bi, leaderLen));
        }
        msg.leader = node;
        msg.prevLogIndex = ByteUtil.readLong(bi);
        msg.prevLogTerm = ByteUtil.readLong(bi);
        msg.leaderCommit = ByteUtil.readLong(bi);
        int entriesLen = ByteUtil.readInt(bi);
        List<LogEntry> list = new ArrayList<>();
        if (entriesLen > 0) {
            for (int i = 0; i < entriesLen; i++) {
                int len = ByteUtil.readInt(bi);
                byte[] bytes = ByteUtil.readBytes(bi, len);
                LogEntry entry = LogEntry.parse(bytes);
                list.add(entry);
            }
        }
        msg.entries = list;
        return msg;
    }

    public RaftMessageType getType() {
        return type;
    }

    public long getTerm() {
        return term;
    }

    public void setTerm(long term) {
        this.term = term;
    }

    public PubKeyNode getLeader() {
        return leader;
    }

    public void setLeader(PubKeyNode leader) {
        this.leader = leader;
    }

    public long getPrevLogIndex() {
        return prevLogIndex;
    }

    public void setPrevLogIndex(long prevLogIndex) {
        this.prevLogIndex = prevLogIndex;
    }

    public long getPrevLogTerm() {
        return prevLogTerm;
    }

    public void setPrevLogTerm(long prevLogTerm) {
        this.prevLogTerm = prevLogTerm;
    }

    public long getLeaderCommit() {
        return leaderCommit;
    }

    public void setLeaderCommit(long leaderCommit) {
        this.leaderCommit = leaderCommit;
    }

    public List<LogEntry> getEntries() {
        return entries;
    }

    public void setEntries(List<LogEntry> entries) {
        this.entries = entries;
    }
}
