package org.bdware.consistency.plugin.raft.algo.message;

import org.bdware.sc.conn.ByteUtil;
import org.bdware.sc.units.PubKeyNode;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;


public class AppendEntriesResponse implements IRaftMessage {
    RaftMessageType type;
    long term;
    long lastLogIndex;
    boolean success;

    PubKeyNode server;

    public AppendEntriesResponse() {
        type = RaftMessageType.AppendEntriesResponse;
    }

    public AppendEntriesResponse(long term, long lastLogIndex, boolean success, PubKeyNode server) {
        type = RaftMessageType.AppendEntriesResponse;
        this.term = term;
        this.lastLogIndex = lastLogIndex;
        this.success = success;
        this.server = server;
    }

    @Override
    public byte[] getBytes() {
        ByteArrayOutputStream msg = new ByteArrayOutputStream();
        ByteUtil.writeLong(msg, term);
        ByteUtil.writeLong(msg, lastLogIndex);
        ByteUtil.writeBoolean(msg, success);
        try {
            if (server != null && !server.pubkey.equals("")) {
                byte[] pubkey = server.pubkey.getBytes();
                ByteUtil.writeInt(msg, pubkey.length);
                msg.write(pubkey);
            } else {
                ByteUtil.writeInt(msg, 0);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return msg.toByteArray();
    }

    @Override
    public IRaftMessage parse(byte[] b) {
        AppendEntriesResponse msg = new AppendEntriesResponse();
        ByteArrayInputStream bi = new ByteArrayInputStream(b);
        msg.term = ByteUtil.readLong(bi);
        msg.lastLogIndex = ByteUtil.readLong(bi);
        msg.success = ByteUtil.readBoolean(bi);
        int serverLen = ByteUtil.readInt(bi);
        PubKeyNode node = new PubKeyNode();
        if (serverLen > 0) {
            node.pubkey = new String(ByteUtil.readBytes(bi, serverLen));
        }
        msg.server = node;
        return msg;
    }

    @Override
    public RaftMessageType getType() {
        return type;
    }

    public long getTerm() {
        return term;
    }

    public void setTerm(long term) {
        this.term = term;
    }

    public boolean isSuccess() {
        return success;
    }

    public void setSuccess(boolean success) {
        this.success = success;
    }

    public PubKeyNode getServer() {
        return server;
    }

    public void setServer(PubKeyNode server) {
        this.server = server;
    }

    public long getLastLogIndex() {
        return lastLogIndex;
    }

    public void setLastLogIndex(long lastLogIndex) {
        this.lastLogIndex = lastLogIndex;
    }
}
