package org.bdware.consistency.plugin.raft.algo.message;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.conn.ByteUtil;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

public class RaftMessageParser {
    IRaftMessage message;
    static Logger LOGGER = LogManager.getLogger(RaftMessageParser.class);

    public RaftMessageParser(IRaftMessage message) {
        this.message = message;
    }

    public byte[] getBytes() {
        ByteArrayOutputStream bo = new ByteArrayOutputStream();
        try {
            ByteUtil.writeInt(bo, message.getType().toInt());
            byte[] bytes = message.getBytes();
            ByteUtil.writeInt(bo, bytes.length);
            bo.write(bytes);
            LOGGER.info("[SEND] " + message.getType());
        } catch (IOException e) {
            e.printStackTrace();
        }
        return bo.toByteArray();
    }

    public static IRaftMessage parse(byte[] b) {
        ByteArrayInputStream bi = new ByteArrayInputStream(b);
        RaftMessageType type = RaftMessageType.fromInt(ByteUtil.readInt(bi));
        int msgLen = ByteUtil.readInt(bi);
        byte[] messageBytes = ByteUtil.readBytes(bi, msgLen);
        IRaftMessage msg;
        switch (type) {
            case RequestVote:
                msg = new RequestVote();
                break;
            case RequestVoteResponse:
                msg = new RequestVoteResponse();
                break;
            case AppendEntries:
                msg = new AppendEntries();
                break;
            case AppendEntriesResponse:
                msg = new AppendEntriesResponse();
                break;
            default:
                msg = new EmptyMessage();
        }
        return msg.parse(messageBytes);
    }

    public static RaftMessageParser setMessage(IRaftMessage message) {
        return new RaftMessageParser(message);
    }
}
