package org.bdware.consistency.plugin.raft.algo.storage;


import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.ByteUtil;
import org.bdware.sc.util.JsonUtil;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;

public class LogEntry {

    private long term;
    private long index;
    private ContractRequest command;


    public long getTerm() {
        return term;
    }

    public void setTerm(long term) {
        this.term = term;
    }

    public long getIndex() {
        return index;
    }

    public void setIndex(long index) {
        this.index = index;
    }

    public ContractRequest getCommand() {
        return command;
    }

    public void setCommand(ContractRequest command) {
        this.command = command;
    }

    public byte[] getBytes() {
        ByteArrayOutputStream bo = new ByteArrayOutputStream();
        ByteUtil.writeLong(bo, term);
        ByteUtil.writeLong(bo, index);
        byte[] dataByte = JsonUtil.toJson(command).getBytes(StandardCharsets.UTF_8);
        try {
            if (dataByte != null && dataByte.length != 0) {
                ByteUtil.writeInt(bo, dataByte.length);
                bo.write(dataByte);
            } else {
                ByteUtil.writeInt(bo, 0);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        return bo.toByteArray();
    }

    public static LogEntry parse(byte[] bytes) {
        LogEntry entry = new LogEntry();
        ByteArrayInputStream bi = new ByteArrayInputStream(bytes);
        entry.term = ByteUtil.readLong(bi);
        entry.index = ByteUtil.readLong(bi);
        int dataLen = ByteUtil.readInt(bi);
        if (dataLen > 0) {
            byte[] dataByte = ByteUtil.readBytes(bi, dataLen);
            entry.command = JsonUtil.fromJson(new String(dataByte, StandardCharsets.UTF_8), ContractRequest.class);
        } else {
            entry.command = null;
        }
        return entry;
    }
}
