package org.bdware.consistency.plugin.raft.algo.util;

import org.bdware.sc.ContractManager;

import java.util.Map;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

public class HeartBeatUtil {
    private static HeartBeatUtil instance;
    Map<TimerTask, ScheduledFuture> recordedFuture;

    private HeartBeatUtil() {
        recordedFuture = new ConcurrentHashMap<>();
    }

    public static HeartBeatUtil getInstance() {
        if (null == instance) {
            instance = new HeartBeatUtil();
        }
        return instance;
    }

    public synchronized void cancel(TimerTask timerTask) {
        try {
            if (recordedFuture.containsKey(timerTask)) {
                ScheduledFuture future = recordedFuture.get(timerTask);
                future.cancel(false);
                recordedFuture.remove(timerTask);
                timerTask.cancel();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public synchronized void schedule(TimerTask timerTask, int delay, int period) {
        try {
            if (!recordedFuture.containsKey(timerTask)) {
                ScheduledFuture<?> future =
                        ContractManager.scheduledThreadPool.scheduleWithFixedDelay(
                                timerTask, delay, period, TimeUnit.MILLISECONDS);
                recordedFuture.put(timerTask, future);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
