package org.bdware.consistency.plugin.raft.algo.storage;

import org.bdware.sc.bean.ContractRequest;
import org.junit.Test;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

public class LogManagerTest {
    @Test
    public void testLogManager() {
        LogManager logManager = new LogManager(System.getProperty("user.dir") + File.separator + "data", 100000);
        List<LogEntry> entries = new ArrayList<>();
        LogEntry entry;
        for (int i = 0; i < 1000; i++) {
            entry = new LogEntry();
            entry.setTerm(i / 10);
            entry.setIndex(i);
            ContractRequest command = new ContractRequest();
            command.setAction(String.format("%20d-%20d", i, i * 100));
            entry.setCommand(command);
            entries.add(entry);
        }
        long idx = logManager.append(entries);
        assert idx == 999;
    }

    @Test
    public void testLogManager2() {
        LogManager logManager = new LogManager(System.getProperty("user.dir") + File.separator + "data", 100000);
        LogEntry entry = logManager.getEntry(386);
        assert entry.getTerm() == 38;
        assert entry.getIndex() == 386;
        assert entry.getCommand().getAction().equals(String.format("%20d-%20d", 386, 386 * 100));
        entry = logManager.getLastEntry();
        assert entry.getTerm() == 99;
        assert entry.getIndex() == 999;
        assert entry.getCommand().getAction().equals(String.format("%20d-%20d", 999, 999 * 100));
        entry = logManager.getEntry(0);
        assert entry.getTerm() == 0;
        assert entry.getIndex() == 0;
        assert entry.getCommand().getAction().equals(String.format("%20d-%20d", 0, 0));
        List<LogEntry> entries = new ArrayList<>();
        for (int i = 1000; i < 1100; i++) {
            entry = new LogEntry();
            entry.setTerm(i / 10);
            entry.setIndex(i);
            ContractRequest command = new ContractRequest();
            command.setAction(String.format("%20d-%20d", i, i * 100));
            entry.setCommand(command);
            entries.add(entry);
        }
        long index = logManager.append(entries);
        assert index == 1099;
    }
    @Test
    public void testLogManager3() {
        LogManager logManager = new LogManager(System.getProperty("user.dir") + File.separator + "data", 100000);
        List<LogEntry> entryList = logManager.getEntriesStartAt(677);
        assert entryList.size() == 1100 - 677;
        assert entryList.get(100).getIndex() == 777;
        logManager.deleteEntriesStartAt(983);
        assert logManager.getLastEntryIndex() == 982;
    }
}
